from __future__ import annotations

import numpy as np
from numpy.typing import ArrayLike


def histogram_bin_edges(data: ArrayLike, bins: int, mode: str, min_count: int = None) -> ArrayLike:
    """
    Determine the bin-edges.

    :param data: The input data.
    :param bins: The number of bins.
    :param mode: The binning mode.
    :param min_count: The minimum number of data-points per bin (only used for mode="uniform").
    """

    if mode == "equal":
        bin_edges = np.linspace(np.min(data), np.max(data), bins + 1)
        bin_edges[0] = np.min(data)
        bin_edges[-1] = np.max(data)
        return bin_edges

    if mode == "log":
        bin_edges = np.logspace(np.log10(np.min(data)), np.log10(np.max(data)), bins + 1)
        bin_edges[0] = np.min(data)
        bin_edges[-1] = np.max(data)
        return bin_edges

    if mode == "uniform":
        if min_count and bins is None:
            if not isinstance(min_count, int):
                raise OSError('"min_count" must be an integer number')
            bins = int(np.floor(float(len(data)) / float(min_count)))

        assert bins is not None

        data = np.array_split(np.sort(data), bins)
        bin_edges = np.empty(bins + 1, dtype=np.float64)
        bin_edges[0] = np.min(data[0])
        bin_edges[-1] = np.max(data[-1])
        for i in range(1, bins):
            bin_edges[i] = 0.5 * (np.max(data[i - 1]) + np.min(data[i]))
        return bin_edges

    if mode == "voronoi":
        mid_points = np.unique(data)
        diff = np.diff(mid_points)
        bin_edges = mid_points + np.hstack((0.5 * diff, 0.5 * diff[-1]))
        return np.hstack((mid_points[0] - 0.5 * diff[0], bin_edges))

    raise OSError("Unknown option")


def histogram_bin_edges_minwidth(bin_edges: ArrayLike, min_width: int) -> ArrayLike:
    r"""
    Merge bin_edges with right-neighbour until each bin has a minimum width.

    :param bin_edges: The bin-edges.
    :param min_width: The minimum bin width.
    :return: The bin-edges.
    """

    while True:
        idx = np.where(np.diff(bin_edges) < min_width)[0]

        if len(idx) == 0:
            return bin_edges

        idx = idx[0]

        if idx + 1 == len(bin_edges) - 1:
            bin_edges = np.hstack((bin_edges[:idx], bin_edges[-1]))
        else:
            j = idx + 1
            k = idx + 2
            bin_edges = np.hstack((bin_edges[:j], bin_edges[k:]))


def histogram_bin_edges_mincount(
    data: ArrayLike, bin_edges: ArrayLike, min_count: int
) -> ArrayLike:
    r"""
    Merge bins with right-neighbour until each bin has a minimum number of data-points.

    :param data: The input data.
    :param bin_edges: The bin-edges.
    :param min_count: The minimum number of data-points per bin.
    :return: The bin-edges.
    """

    assert isinstance(min_count, int)

    while True:
        count, _ = np.histogram(data, bins=bin_edges, density=False)

        idx = np.where(count < min_count)[0]

        if len(idx) == 0:
            return bin_edges

        idx = idx[0]

        if idx + 1 == len(count):
            bin_edges = np.hstack((bin_edges[:idx], bin_edges[-1]))
        else:
            j = idx + 1
            k = idx + 2
            bin_edges = np.hstack((bin_edges[:j], bin_edges[k:]))


def histogram_bin_edges_integer(bin_edges: ArrayLike) -> ArrayLike:
    r"""
    Merge bins not encompassing an integer with the preceding bin.
    For example: a bin with edges ``[1.1, 1.9]`` is removed, but ``[0.9, 1.1]`` is not removed.

    :param bin_edges: The bin-edges.
    :return: The bin-edges.
    """

    bin_edges = np.array(bin_edges)
    assert bin_edges.size > 1

    i = np.where(np.diff(np.floor(bin_edges)) >= 1)[0]

    if i[0] > 0:
        i[0] = 0

    i = list(i) + [bin_edges.size - 1]

    return bin_edges[i]



project = "enstat"
copyright = "2021, Tom de Geus"
author = "Tom de Geus"
autodoc_type_aliases = {"Iterable": "Iterable", "ArrayLike": "ArrayLike"}
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
html_theme = "furo"
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx_mdinclude",
]



import unittest

import numpy as np

import enstat


class Test_norm(unittest.TestCase):
    """
    Histogram normalisation.
    """

    def test_density(self):
        data = [0, 0, 0, 1, 1, 2]
        bin_edges = [-0.5, 0.5, 1.5, 2.5]
        c, _ = np.histogram(data, bins=bin_edges, density=False)
        p, _ = np.histogram(data, bins=bin_edges, density=True)

        hist = enstat.histogram.from_data(data, bin_edges=bin_edges)

        self.assertTrue(np.allclose(hist.count, c))
        self.assertTrue(np.allclose(hist.density, p))

    def test_density_random(self):
        data = np.random.random([1001])

        for bin_edges in [
            np.logspace(-5, 1, 1000),
            np.linspace(0, 1, 1000),
            np.sort(np.random.random([10])),
        ]:
            c, _ = np.histogram(data, bins=bin_edges, density=False)
            p, _ = np.histogram(data, bins=bin_edges, density=True)
            hist = enstat.histogram.from_data(data, bin_edges=bin_edges, bound_error="ignore")
            self.assertTrue(np.allclose(hist.count, c))
            self.assertTrue(np.allclose(hist.density, p))

    def test_density_random_binning(self):
        data = np.random.random([1001])

        for mode in ["equal", "log", "uniform", "voronoi"]:
            hist = enstat.histogram.from_data(data, bins=50, mode=mode)
            c, _ = np.histogram(data, bins=hist.bin_edges, density=False)
            p, _ = np.histogram(data, bins=hist.bin_edges, density=True)
            self.assertTrue(np.all(hist.count == c))
            self.assertTrue(np.allclose(hist.density, p))

    def test_bound_error_norm(self):
        data = np.random.random([1001])

        r = np.random.random([30])
        bin_edges = np.min(data) + np.cumsum(0.1 * r)
        bin_edges = bin_edges[bin_edges < np.max(data)]
        bin_edges_full = np.hstack((np.min(data), bin_edges, np.max(data)))

        hist = enstat.histogram.from_data(data, bin_edges=bin_edges, bound_error="norm")
        hist_full = enstat.histogram.from_data(data, bin_edges=bin_edges_full, bound_error="raise")

        self.assertTrue(np.all(hist.count == hist_full.count[1:-1]))
        self.assertTrue(np.allclose(hist.p, hist_full.p[1:-1]))

    def test_mid(self):
        bin_edges = [-0.5, 0.5, 1.5, 2.5]
        mid = [0, 1, 2]

        hist = enstat.histogram(bin_edges=bin_edges)

        self.assertEqual(mid, hist.x.tolist())


class Test_accumulate(unittest.TestCase):
    """
    Histogram accumulation.
    """

    def test_accumulate(self):
        data = np.random.random([100, 100])
        bin_edges = np.linspace(0, 1, 21)

        c, _ = np.histogram(data.ravel(), bins=bin_edges, density=False)
        p, _ = np.histogram(data.ravel(), bins=bin_edges, density=True)

        hist = enstat.histogram(bin_edges=bin_edges)

        for i in range(data.shape[0]):
            hist += data[i, :]

        self.assertEqual(hist.count_left, 0)
        self.assertEqual(hist.count_right, 0)

        self.assertTrue(np.all(hist.count == c))
        self.assertTrue(np.allclose(hist.density, p))

    def test_accumulate_bin_edges(self):
        data = np.random.random([100, 100])

        for mode in ["equal", "log", "uniform", "voronoi"]:
            hist = enstat.histogram.from_data(data.ravel(), bins=51, mode=mode)
            bin_edges = np.copy(hist.bin_edges)

            c, _ = np.histogram(data.ravel(), bins=bin_edges, density=False)
            p, _ = np.histogram(data.ravel(), bins=bin_edges, density=True)

            hist = enstat.histogram(bin_edges=bin_edges)

            for i in range(data.shape[0]):
                hist += data[i, :]

            self.assertEqual(hist.count_left, 0)
            self.assertEqual(hist.count_right, 0)

            self.assertTrue(np.all(hist.count == c))
            self.assertTrue(np.allclose(hist.density, p))


class Test_strip(unittest.TestCase):
    """
    Strip empty bins.
    """

    def test_lstrip(self):
        data = np.random.random(100)
        bin_edges = np.linspace(-1, 2, 25)
        count, _ = np.histogram(data, bins=bin_edges, density=False)

        hist = enstat.histogram.from_data(data, bin_edges=bin_edges)
        hist.lstrip()

        self.assertTrue(np.allclose(hist.bin_edges, bin_edges[8:]))
        self.assertTrue(np.all(hist.count == count[8:]))

    def test_rstrip(self):
        data = np.random.random(100)
        bin_edges = np.linspace(-1, 2, 25)
        count, _ = np.histogram(data, bins=bin_edges, density=False)

        hist = enstat.histogram.from_data(data, bin_edges=bin_edges)
        hist.rstrip()

        self.assertTrue(np.allclose(hist.bin_edges, bin_edges[:-8]))
        self.assertTrue(np.all(hist.count == count[:-8]))

    def test_strip(self):
        data = np.random.random(100)
        bin_edges = np.linspace(-1, 2, 25)
        count, _ = np.histogram(data, bins=bin_edges, density=False)

        hist = enstat.histogram.from_data(data, bin_edges=bin_edges)
        hist.strip()

        self.assertTrue(np.allclose(hist.bin_edges, bin_edges[8:-8]))
        self.assertTrue(np.all(hist.count == count[8:-8]))


class Test_merge(unittest.TestCase):
    """
    Merge bins.
    """

    def test_merge_right(self):
        bin_edges = [0, 1, 2, 3, 4, 5]
        count = [1, 2, 1, 2, 3]

        hist = enstat.histogram.restore(bin_edges=bin_edges, count=count)
        hist.merge_right([0, 2])

        self.assertEqual([0, 2, 4, 5], hist.bin_edges.tolist())
        self.assertEqual([3, 3, 3], hist.count.tolist())

    def test_merge_right_first(self):
        bin_edges = [0, 1, 2, 3, 4, 5]
        count = [1, 2, 1, 2, 3]

        hist = enstat.histogram.restore(bin_edges=bin_edges, count=count)
        hist.merge_right([0])

        self.assertEqual([0, 2, 3, 4, 5], hist.bin_edges.tolist())
        self.assertEqual([3, 1, 2, 3], hist.count.tolist())

    def test_merge_right_last(self):
        bin_edges = [0, 1, 2, 3, 4, 5]
        count = [1, 2, 1, 2, 3]

        hist = enstat.histogram.restore(bin_edges=bin_edges, count=count)
        hist.merge_right([-1])

        self.assertEqual(bin_edges, hist.bin_edges.tolist())
        self.assertEqual(count, hist.count.tolist())

    def test_merge_left(self):
        bin_edges = [0, 1, 2, 3, 4, 5]
        count = [1, 2, 1, 2, 3]

        hist = enstat.histogram.restore(bin_edges=bin_edges, count=count)
        hist.merge_left([0, 2])

        self.assertEqual([0, 1, 3, 4, 5], hist.bin_edges.tolist())
        self.assertEqual([1, 3, 2, 3], hist.count.tolist())

    def test_merge_left_first(self):
        bin_edges = [0, 1, 2, 3, 4, 5]
        count = [1, 2, 1, 2, 3]

        hist = enstat.histogram.restore(bin_edges=bin_edges, count=count)
        hist.merge_left([0])

        self.assertEqual(bin_edges, hist.bin_edges.tolist())
        self.assertEqual(count, hist.count.tolist())

    def test_merge_left_last(self):
        bin_edges = [0, 1, 2, 3, 4, 5]
        count = [1, 2, 1, 2, 3]

        hist = enstat.histogram.restore(bin_edges=bin_edges, count=count)
        hist.merge_left([-1])

        self.assertEqual([0, 1, 2, 3, 5], hist.bin_edges.tolist())
        self.assertEqual([1, 2, 1, 5], hist.count.tolist())

    def test_as_integer(self):
        bin_edges = [0, 1, 2, 2.1, 3, 3.1, 4, 5]
        count = [1, 2, 1, 1, 1, 2, 3]

        hist = enstat.histogram.restore(bin_edges=bin_edges, count=count)
        hist.as_integer()

        self.assertEqual([0, 1, 2, 3, 4, 5], hist.bin_edges.tolist())
        self.assertEqual([1, 2, 2, 3, 3], hist.count.tolist())

    def test_as_integer_first(self):
        bin_edges = [0, 0.9, 1, 2, 3, 4, 5]
        count = [1, 2, 1, 1, 2, 3]

        hist = enstat.histogram.restore(bin_edges=bin_edges, count=count)
        hist.as_integer()

        self.assertEqual([0, 1, 2, 3, 4, 5], hist.bin_edges.tolist())
        self.assertEqual([3, 1, 1, 2, 3], hist.count.tolist())

    def test_as_integer_last(self):
        bin_edges = [0, 1, 2, 3, 4, 4.5, 5]
        count = [1, 2, 1, 2, 1, 3]

        hist = enstat.histogram.restore(bin_edges=bin_edges, count=count)
        hist.as_integer()

        self.assertEqual([0, 1, 2, 3, 4, 5], hist.bin_edges.tolist())
        self.assertEqual([1, 2, 1, 2, 4], hist.count.tolist())


class Test_squash(unittest.TestCase):
    """
    Squash ``n`` bins into one.
    """

    def test_squash_fit(self):
        bin_edges = [0, 1, 2, 3, 4, 6]
        count = [1, 2, 1, 2, 3]

        hist = enstat.histogram.restore(bin_edges=bin_edges, count=count)
        hist.squash(2)

        self.assertEqual(0, hist.count_left)
        self.assertEqual(0, hist.count_right)

        self.assertEqual([3, 3, 3], hist.count.tolist())
        self.assertEqual([0, 2, 4, 6], hist.bin_edges.tolist())

    def test_squash_nofit(self):
        bin_edges = [0, 1, 2, 3, 4, 6]
        count = [1, 2, 1, 2, 3]

        hist = enstat.histogram.restore(bin_edges=bin_edges, count=count)
        hist.squash(3)

        self.assertEqual(0, hist.count_left)
        self.assertEqual(0, hist.count_right)

        self.assertEqual([4, 5], hist.count.tolist())
        self.assertEqual([0, 3, 6], hist.bin_edges.tolist())

        hist = enstat.histogram.restore(bin_edges=bin_edges, count=count)
        hist.squash(4)

        self.assertEqual([6, 3], hist.count.tolist())
        self.assertEqual([0, 4, 6], hist.bin_edges.tolist())

        hist = enstat.histogram.restore(bin_edges=bin_edges, count=count)
        hist.squash(5)

        self.assertEqual([9], hist.count.tolist())
        self.assertEqual([0, 6], hist.bin_edges.tolist())

        hist = enstat.histogram.restore(bin_edges=bin_edges, count=count)
        hist.squash(6)

        self.assertEqual([9], hist.count.tolist())
        self.assertEqual([0, 6], hist.bin_edges.tolist())

    def test_strip_squash(self):
        bin_edges = [-1, 0, 1, 2, 3, 4, 6, 7]
        count = [0, 1, 2, 1, 2, 3, 0]

        hist = enstat.histogram.restore(bin_edges=bin_edges, count=count)
        hist.strip()
        hist.squash(2)

        self.assertEqual([3, 3, 3], hist.count.tolist())
        self.assertEqual([0, 2, 4, 6], hist.bin_edges.tolist())


class Test_histogram_bin_edges_integer(unittest.TestCase):
    """
    Bin edges: integer
    """

    def test_front(self):
        a = [0, 0.5, 1.5, 2.5]
        b = [0, 1.5, 2.5]
        self.assertTrue(np.allclose(enstat.detail.histogram_bin_edges_integer(a), b))

    def test_middle(self):
        a = [0, 1.5, 1.6, 2.5]
        b = [0, 1.6, 2.5]
        self.assertTrue(np.allclose(enstat.detail.histogram_bin_edges_integer(a), b))

    def test_back(self):
        a = [0, 1.5, 2.5, 2.6]
        b = [0, 1.5, 2.6]
        self.assertTrue(np.allclose(enstat.detail.histogram_bin_edges_integer(a), b))


class Test_histogram_bin_edges_voronoi(unittest.TestCase):
    """
    Bin edges: Voronoi
    """

    def test_integer(self):
        data = np.array([0, 0, 1, 1, 1, 2, 2, 3, 4, 5, 6])
        hist = enstat.histogram.from_data(data, bins=10, mode="voronoi")

        self.assertEqual(hist.count_left, 0)
        self.assertEqual(hist.count_right, 0)

        self.assertTrue(
            np.allclose(hist.bin_edges, np.array([-0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5]))
        )

    def test_integer2(self):
        data = np.array([8, 8, 1, 1, 1, 2, 2, 3, 4, 5, 6])
        hist = enstat.histogram.from_data(data, bins=10, mode="voronoi")

        self.assertEqual(hist.count_left, 0)
        self.assertEqual(hist.count_right, 0)

        self.assertTrue(np.allclose(hist.bin_edges, np.array([0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 7, 9])))


class Test_histogram_restore(unittest.TestCase):
    """
    Restore a previous result.
    """

    def test_basic(self):
        data = np.random.random([1001])

        r = np.random.random([30])
        bin_edges = np.min(data) + np.cumsum(0.1 * r)
        bin_edges = bin_edges[bin_edges < np.max(data)]
        bin_edges_full = np.hstack((np.min(data), bin_edges, np.max(data)))

        hist = enstat.histogram.from_data(data, bin_edges=bin_edges, bound_error="norm")
        hist_full = enstat.histogram.from_data(data, bin_edges=bin_edges_full, bound_error="raise")
        restored = enstat.histogram.restore(**dict(hist))

        self.assertTrue(np.all(hist.count == hist_full.count[1:-1]))
        self.assertTrue(np.allclose(hist.p, hist_full.p[1:-1]))

        self.assertTrue(np.all(hist.count == restored.count))
        self.assertTrue(np.allclose(hist.p, restored.p))


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import numpy as np
from numpy.typing import ArrayLike
from numpy.typing import DTypeLike

from . import detail  # noqa: F401
from . import mean  # noqa: F401
from ._version import version  # noqa: F401
from ._version import version_tuple  # noqa: F401


class scalar:
    r"""
    Ensemble average of a scalar.
    Add samples to it using :py:func:`scalar.add_sample`.
    The mean, variance, and standard deviation can be obtained at any time.
    They are derived from the following members:

    *   :py:attr:`scalar.first`: Sum of the first statistical moment.
    *   :py:attr:`scalar.second`: Sum of the second statistical moment.
    *   :py:attr:`scalar.norm`: Number of samples.
    """

    def __init__(self):
        self.first = 0.0
        self.second = 0.0
        self.norm = 0.0

    def __iter__(self):
        yield "first", self.first
        yield "second", self.second
        yield "norm", self.norm

    @classmethod
    def restore(cls, first: float = 0, second: float = 0, norm: float = 0):
        """
        Restore previous data.

        :param float first: Sum of the first moment.
        :param float second: Sum of the second moment.
        :param int norm: Number of samples.
        """

        ret = cls()
        ret.first = first
        ret.second = second
        ret.norm = norm
        return ret

    def __add__(self, datum: float):
        self.add_sample(datum)
        return self

    def add_sample(self, datum: float):
        """
        Add a sample.
        Internally changes the sums of the first and second statistical moments and normalisation.

        :param datum: Sample.
        """

        datum = np.array(datum)
        self.first += np.sum(datum)
        self.second += np.sum(datum**2)
        self.norm += datum.size

    def mean(self) -> float:
        r"""
        Current mean.
        Samples can be added afterwards without any problems.

        :return: Mean.
        """

        if self.norm == 0:
            return np.NaN

        return self.first / self.norm

    def variance(self) -> float:
        r"""
        Current variance.
        Samples can be added afterwards without any problems.

        :return: Variance.
        """

        if self.norm <= 1:
            return np.NaN

        d = (self.second / self.norm - (self.first / self.norm) ** 2) * self.norm
        n = self.norm - 1

        return d / n

    def std(self) -> float:
        r"""
        Current standard deviation.
        Samples can be added afterwards without any problems.

        :return: Standard deviation.
        """

        return np.sqrt(self.variance())


class static:
    r"""
    Ensemble average of an nd-array (of same size for all samples).
    Add samples to it using :py:func:`static.add_sample`.
    The mean, variance, and standard deviation can be obtained at any time.
    They are derived from the following members:

    *   :py:attr:`static.first`: Sum of the first statistical moment.
    *   :py:attr:`static.second`: Sum of the second statistical moment.
    *   :py:attr:`static.norm`: Number of samples.

    Furthermore, the following members are available:

    *   :py:attr:`static.shape`: Shape of the data.
    *   :py:attr:`static.size`: Size of the data (= prod(shape)).

    :param compute_variance:
        If set ``False`` no second moment will be computed (making things slightly faster).
        In that case, the variance an standard deviation will not be available.

    :param shape:
        The shape of the data.
        If not specified it is determined form the first sample.

    :param dtype:
        The type of the data.
        If not specified it is determined form the first sample.
    """

    def __init__(
        self,
        compute_variance: bool = True,
        shape: tuple[int] = None,
        dtype: DTypeLike = np.float64,
    ):
        self.compute_variance = compute_variance
        self.norm = None
        self.first = None
        self.second = None

        if shape is None:
            return

        self.norm = np.zeros(shape, np.int64)
        self.first = np.zeros(shape, dtype)
        if compute_variance:
            self.second = np.zeros(shape, dtype)

    def __iter__(self):
        yield "first", self.first
        yield "second", self.second
        yield "norm", self.norm

    @classmethod
    def restore(
        cls,
        first: ArrayLike = None,
        second: ArrayLike = None,
        norm: ArrayLike = None,
    ):
        """
        Restore previous data.

        :param first: Continued computation: Sum of the first moment.
        :param second: Continued computation: Sum of the second moment.
        :param norm: Continued computation: Number of samples (integer).
        """

        if norm is not None:
            assert first is not None
            assert first.shape == norm.shape

            if second is not None:
                assert second.shape == norm.shape

        ret = cls()
        ret.first = first
        ret.second = second
        ret.norm = norm
        return ret

    def _allocate(self, shape, dtype):
        self.norm = np.zeros(shape, np.int64)
        self.first = np.zeros(shape, dtype)
        if self.compute_variance:
            self.second = np.zeros(shape, dtype)

    @property
    def dtype(self):
        """
        The type of the data.
        """
        return self.first.dtype

    @property
    def shape(self):
        """
        The shape of the data.
        """
        return self.first.shape

    @property
    def size(self):
        """
        The size of the data.
        """
        return np.prod(self.first.shape)

    def ravel(self) -> scalar:
        r"""
        Return as :py:class:`scalar`: all entries are summed.

        :return: Ensemble average.
        """

        return scalar.restore(
            first=np.sum(self.first),
            second=np.sum(self.second),
            norm=np.sum(self.norm),
        )

    def __add__(self, data: ArrayLike):
        self.add_sample(data)
        return self

    def add_sample(self, data: ArrayLike, mask: ArrayLike = None):
        r"""
        Add a sample.
        Internally changes the sums of the first and second statistical moments and normalisation.

        :param data: The sample.
        :param mask: Mask entries (boolean array).
        """

        if self.first is None:
            self._allocate(data.shape, data.dtype)

        # masked data
        if mask is not None:
            incl = np.logical_not(mask)
            self.norm[incl] += 1
            self.first[incl] += data[incl]
            if self.compute_variance:
                self.second[incl] += data[incl] ** 2
            return

        # unmasked data
        self.norm += 1
        self.first += data
        if self.compute_variance:
            self.second += data**2

    def add_point(self, datum: float | int, index: int):
        """
        Add a single point. Note that::

            ensemble.add_point(datum, index)

        Is equivalent to::

            data = np.empty(ensemble.shape)
            mask = np.ones(ensemble.shape, dtype=bool)
            data[index] = datum
            mask[index] = False
            ensemble.add_sample(data, mask)

        (but faster).
        """

        if self.first is None:
            raise OSError("shape should be pre-specified")

        self.norm[index] += 1
        self.first[index] += datum

        if self.compute_variance:
            self.second[index] += datum**2

    def mean(self, min_norm: int = 1) -> ArrayLike:
        r"""
        Current mean.
        Samples can be added afterwards without any problems.

        :param min_norm: Minimum number of samples to consider as value output.
        :return: Mean.
        """

        if self.norm is None:
            return None

        n = np.where(self.norm > 0, self.norm, 1)
        ret = self.first / n
        return np.where(self.norm >= min_norm, ret, np.NaN)

    def variance(self, min_norm: int = 2) -> ArrayLike:
        r"""
        Current variance.
        Samples can be added afterwards without any problems.

        :param min_norm: Minimum number of samples to consider as value output.
        :return: Variance.
        """

        if self.norm is None:
            return None

        assert self.compute_variance
        n = np.where(self.norm > 1, self.norm, 2)
        ret = (self.second / n - (self.first / n) ** 2) * n / (n - 1)
        return np.where(self.norm >= min_norm, ret, np.NaN)

    def std(self, min_norm: int = 2) -> ArrayLike:
        r"""
        Current standard deviation.
        Samples can be added afterwards without any problems.

        :param min_norm: Minimum number of samples to consider as value output.
        :return: Standard deviation.
        """

        if self.norm is None:
            return None

        return np.sqrt(self.variance(min_norm))


def _expand_array1d(data, size):
    tmp = np.zeros((size), data.dtype)
    tmp[: data.size] = data
    return tmp


class dynamic1d(static):
    r"""
    Ensemble average of an 1d-array (which grows depending of the size of the samples).
    Add samples to it using :py:func:`dynamic1d.add_sample`.
    The mean, variance, and standard deviation can be obtained at any time.
    Also the sums of the first and statistical moments, as well as the number of samples can be
    obtained at any time.

    Continue an old average by specifying:

    :param compute_variance:
        If set ``False`` no second moment will be computed.
        In that case, the variance an standard deviation will not be available.

    :param size:
        The initial size of the data.
        If not specified it is determined form the first sample.

    :param dtype:
        The type of the data.
        If not specified it is determined form the first sample.
    """

    def __init__(
        self,
        compute_variance: bool = True,
        size: int = None,
        dtype: DTypeLike = np.float64,
    ):
        super().__init__(
            compute_variance=compute_variance,
            shape=(size,) if size is not None else None,
            dtype=dtype,
        )

    def _expand(self, size: int):
        if size <= self.first.size:
            return

        self.norm = _expand_array1d(self.norm, size)
        self.first = _expand_array1d(self.first, size)

        if self.compute_variance:
            self.second = _expand_array1d(self.second, size)

    def __add__(self, data: ArrayLike):
        self.add_sample(data)
        return self

    def add_sample(self, data: ArrayLike):
        assert data.ndim == 1

        if self.first is None:
            super().add_sample(data)
            return

        self._expand(data.size)
        self.norm[: data.size] += 1
        self.first[: data.size] += data

        if self.compute_variance:
            self.second[: data.size] += data**2

    def add_point(self, datum: float | int, index: int):
        if self.first is None:
            self._allocate(index + 1, type(datum))
        else:
            self._expand(index + 1)

        return super().add_point(datum, index)


class histogram:
    """
    Histogram.
    One can add samples to it using :py:func:`Histogram.add_sample`.

    :param bin_edges: The bin-edges.
    :param right: Whether the bin includes the right edge (or the left edge) see numpy.digitize.
    :param bound_error: What to do if a sample falls out of the bin range:

        - ``"raise"``: raise an error
        - ``"ignore"``: ignore the data that are out of range
        - ``"norm"``: change the normalisation of the density
    """

    def __init__(
        self,
        bin_edges: ArrayLike,
        right: bool = False,
        bound_error: str = "raise",
    ):
        assert np.all(np.diff(bin_edges) > 0) or np.all(np.diff(bin_edges) < 0)

        self.right = right
        self.bin_edges = np.array(bin_edges).astype(np.float64)
        self.bound_error = bound_error
        self.count_left = 0
        self.count_right = 0
        self.bound_left = bin_edges[0]
        self.bound_right = bin_edges[-1]
        self.count = np.zeros((len(bin_edges) - 1), np.uint64)

    def __iter__(self):
        yield "bin_edges", self.bin_edges
        yield "count", self.count
        yield "count_left", self.count_left
        yield "count_right", self.count_right
        yield "bound_left", self.bound_left
        yield "bound_right", self.bound_right
        yield "right", self.right
        yield "bound_error", self.bound_error

    @classmethod
    def restore(
        cls,
        bin_edges: ArrayLike,
        count: ArrayLike,
        count_left: int = 0,
        count_right: int = 0,
        bound_left: float = None,
        bound_right: float = None,
        bound_error: str = "raise",
        right: bool = False,
    ):
        """
        Restore from a previous result::

            hist = enstat.histogram...
            state = dict(hist)

            restored = enstat.histogram.from_histogram(**state)

        :param bin_edges: The bin-edges.
        :param count: The count.
        :param count_left: Number of items below the left bound.
        :param count_right: Number of items above the right bound.
        :param bound_left: The minimum value below the left bound.
        :param bound_right: The maximum value above the right bound.
        :param bound_error: What to do if a sample falls out of the bin range.
        :param right: Whether the bin includes the right edge (or the left edge) see numpy.digitize.
        """

        ret = cls(
            bin_edges=bin_edges,
            right=right,
            bound_error=bound_error,
        )

        assert len(count) == len(bin_edges) - 1
        ret.count = np.array(count).astype(np.uint64)

        ret.count_left = count_left
        ret.count_right = count_right

        if bound_left is not None:
            assert bound_left <= bin_edges[0]
            ret.bound_left = bound_left

        if bound_right is not None:
            assert bound_right >= bin_edges[-1]
            ret.bound_right = bound_right

        return ret

    @classmethod
    def from_data(
        cls,
        data: ArrayLike,
        bins: int = None,
        mode: str = "equal",
        min_count: int = None,
        min_width: float = None,
        integer: bool = False,
        bin_edges: ArrayLike = None,
        bound_error: str = "raise",
    ):
        r"""
        Construct a histogram from data.

        :param data: Data (flattened).
        :param bins: Number of bins.
        :param mode:
            Mode with which to compute the bin-edges.

            -   ``'equal'``: each bin has equal width.
            -   ``'log'``: logarithmic spacing.
            -   ``'uniform'``: uniform number of data-points per bin.
            -   ``'voronoi'``: each bin is the region between two adjacent data-points.

        :param min_count: Minimum number of data-points per bin.
        :param min_width: Minimum width of a bin.

        :param integer:
            If ``True``, bins not encompassing an integer are removed
            (e.g. a bin with edges ``[1.1, 1.9]`` is removed, but ``[0.9, 1.1]`` is not removed).

        :param bin_edges: Specify the bin-edges (overrides ``bins`` and ``mode``).

        :param bound_error: What to do if a sample falls out of the bin range:

            - ``"raise"``: raise an error
            - ``"ignore"``: ignore the data that are out of range
            - ``"norm"``: change the normalisation of the density

        :return: The :py:class:`Histogram` object.
        """

        if hasattr(bins, "__len__"):
            raise OSError("Only the number of bins can be specified")

        if bin_edges is None:
            bin_edges = detail.histogram_bin_edges(data, bins, mode, min_count)

        if min_count is not None:
            bin_edges = detail.histogram_bin_edges_mincount(data, bin_edges, min_count)

        if min_width is not None:
            bin_edges = detail.histogram_bin_edges_minwidth(bin_edges, min_width)

        if integer:
            bin_edges = detail.histogram_bin_edges_integer(bin_edges)

        bin_edges = np.array(bin_edges)
        bin_edges[0] -= np.finfo(bin_edges.dtype).eps
        bin_edges[-1] += np.finfo(bin_edges.dtype).eps
        return cls(bin_edges, right=True, bound_error=bound_error).add_sample(data)

    def strip(self, min_count: int = 0):
        """
        Strip the histogram of empty bins to the left and the right.

        :param min_count: The minimum count for a bin to be considered non-empty.
        """

        self.lstrip(min_count)
        self.rstrip(min_count)
        return self

    def lstrip(self, min_count: int = 0):
        """
        Strip the histogram of empty bins to the left.

        :param min_count: The minimum count for a bin to be considered non-empty.
        """

        i = np.argmax(self.count > min_count)
        self.count = self.count[i:]
        self.bin_edges = self.bin_edges[i:]
        return self

    def rstrip(self, min_count: int = 0):
        """
        Strip the histogram of empty bins to the right.

        :param min_count: The minimum count for a bin to be considered non-empty.
        """

        i = len(self.count) - np.argmax(self.count[::-1] > min_count)
        self.count = self.count[:i]
        self.bin_edges = self.bin_edges[: i + 1]
        return self

    def interp(self, bin_edges: ArrayLike):
        """
        Interpolate the histogram to a new set of bin-edges.

        :param bin_edges: The new bin-edges.
        """

        assert np.all(np.diff(bin_edges) > 0) or np.all(np.diff(bin_edges) < 0)

        m = 0.5 * (bin_edges[:-1] + bin_edges[1:])
        self.count = np.interp(m, self.x, self.count)
        self.bin_edges = bin_edges
        return self

    def squash(self, n: int):
        """
        Squash the histogram by combining ``n`` sequential bins into one
        (the last bin may be smaller).

        :param n: Number of bins to group.
        """

        if n >= self.count.size:
            self.count = np.sum(self.count).reshape(1)
            self.bin_edges = self.bin_edges[[0, -1]]
            return self

        m = self.count.size // n
        s = m * n
        self.count = np.hstack((self.count[:s].reshape((m, n)).sum(axis=1), np.sum(self.count[s:])))

        if len(self.bin_edges) // n != self.count.size + 1:
            self.bin_edges = np.hstack((self.bin_edges[::n], self.bin_edges[-1]))
        else:
            self.bin_edges = self.bin_edges[::n]

        return self

    def merge_right(self, index: ArrayLike):
        """
        Merge the bins to the right of ``index`` into ``index``.

        :param index: The indices of the bin to merge into.
        """

        if isinstance(index, int):
            index = [index]

        index = np.array(index)
        index = np.sort(np.where(index < 0, index + self.count.size, index))

        if index[-1] == self.count.size - 1:
            index = index[:-1]

        self.count[index] += self.count[index + 1]
        self.count = np.delete(self.count, index + 1)
        self.bin_edges = np.delete(self.bin_edges, index + 1)
        return self

    def merge_left(self, index: ArrayLike):
        """
        Merge the bins to the left of ``index`` into ``index``.

        :param index: The indices of the bin to merge into.
        """

        if isinstance(index, int):
            index = [index]

        index = np.array(index)
        index = np.sort(np.where(index < 0, index + self.count.size, index))

        if index[0] == 0:
            index = index[1:]

        self.count[index - 1] += self.count[index]
        self.count = np.delete(self.count, index)
        self.bin_edges = np.delete(self.bin_edges, index)
        return self

    def as_integer(self):
        """
        Merge bins not encompassing an integer with the preceding bin.
        For example: a bin with edges ``[1.1, 1.9]`` is removed, but ``[0.9, 1.1]`` is not removed.
        """

        assert self.bin_edges.size > 1

        merge = np.argwhere(np.diff(np.floor(self.bin_edges)) == 0).ravel()
        self.merge_right(merge)
        self.bin_edges = np.floor(self.bin_edges)
        self.right = False

    def __add__(self, data: ArrayLike):
        self.add_sample(data)
        return self

    def add_sample(self, data: ArrayLike):
        """
        Add a sample to the histogram.
        You can also use the ``+`` operator.
        """

        bin = np.digitize(data, self.bin_edges, self.right) - 1

        left = bin < 0
        right = bin >= self.count.size
        nleft = np.sum(left)
        nright = np.sum(right)
        self.count_left += nleft
        self.count_right += nright

        if nleft:
            self.bound_left = min([self.bound_left, np.min(data[left])])

        if nright:
            self.bound_right = max([self.bound_right, np.max(data[right])])

        if self.bound_error == "raise" and (nleft or nright):
            raise ValueError("Data out of bin range")

        bin = bin[np.logical_and(~left, ~right)]
        self.count += np.bincount(bin, minlength=self.count.size).astype(np.uint64)
        return self

    @property
    def x(self) -> ArrayLike:
        """
        The bin centers.
        """

        return 0.5 * (self.bin_edges[:-1] + self.bin_edges[1:])

    @property
    def density(self) -> ArrayLike:
        """
        The probability density function at the bin.
        """

        count = self.count.astype(np.float64)
        norm = np.sum(count)

        if self.bound_error == "norm":
            norm += self.count_left + self.count_right

        return count / (np.diff(self.bin_edges) * norm)

    @property
    def p(self) -> ArrayLike:
        """
        The probability density function at the bin.
        """

        return self.density

    @property
    def plot(self) -> tuple[ArrayLike, ArrayLike]:
        """
        Alias for ``(x, density)``.
        """

        return (self.x, self.density)


if __name__ == "__main__":
    pass



import unittest
from collections import defaultdict

import numpy as np

import enstat.mean


class Test_mean(unittest.TestCase):
    """
    tests
    """

    def test_scalar(self):
        """
        Basic test of "mean" and "std" using a random sample.
        """

        average = enstat.scalar()

        average += np.array(1.0)

        self.assertFalse(np.isnan(average.mean()))
        self.assertTrue(np.isnan(average.std()))

        average += np.array(1.0)

        self.assertFalse(np.isnan(average.mean()))
        self.assertFalse(np.isnan(average.std()))

    def test_scalar_division(self):
        """
        Check for zero division.
        """

        average = enstat.scalar()

        a = np.random.random(50 * 20).reshape(50, 20)

        for i in range(a.shape[0]):
            average += a[i, :]

        self.assertTrue(np.isclose(average.mean(), np.mean(a)))
        self.assertTrue(np.isclose(average.std(), np.std(a), rtol=1e-3))

    def test_static(self):
        """
        Basic test of "mean" and "std" using a random sample.
        """

        average = enstat.static()

        a = np.random.random(35 * 50 * 20).reshape(35, 50, 20)

        for i in range(a.shape[0]):
            average += a[i, :, :]

        self.assertTrue(np.allclose(average.mean(), np.mean(a, axis=0)))
        self.assertTrue(np.allclose(average.std(), np.std(a, axis=0), rtol=5e-1, atol=1e-3))
        self.assertTrue(average.shape == a.shape[1:])
        self.assertTrue(average.size == np.prod(a.shape[1:]))

    def test_static_ravel(self):
        """
        Like :py:func:`test_static` but with a test of `ravel`.
        """

        arraylike = enstat.static()
        scalar = enstat.scalar()

        a = np.random.random(35 * 50 * 20).reshape(35, 50, 20)

        for i in range(a.shape[0]):
            arraylike += a[i, :, :]
            scalar += a[i, :, :]

        flat = arraylike.ravel()

        self.assertTrue(np.allclose(flat.mean(), np.mean(a)))
        self.assertTrue(np.allclose(flat.std(), np.std(a), rtol=5e-1, atol=1e-3))
        self.assertTrue(np.allclose(flat.mean(), scalar.mean()))
        self.assertTrue(np.allclose(flat.std(), scalar.std(), rtol=5e-1, atol=1e-3))

    def test_static_division(self):
        """
        Check for zero division.
        """

        average = enstat.static()

        average += np.array([1.0])

        self.assertFalse(np.isnan(average.mean()))
        self.assertTrue(np.isnan(average.std()))

        average += np.array([1.0])

        self.assertFalse(np.isnan(average.mean()))
        self.assertFalse(np.isnan(average.std()))

    def test_static_mask(self):
        """
        Mask part of the data.
        """

        average = enstat.static()

        a = np.random.random(35 * 50 * 20).reshape(35, 50, 20)
        m = np.random.random(35 * 50 * 20).reshape(35, 50, 20) > 0.8

        for i in range(a.shape[0]):
            average.add_sample(a[i, :, :], m[i, :, :])

        self.assertTrue(
            np.isclose(
                np.sum(average.first) / np.sum(average.norm),
                np.mean(a[np.logical_not(m)]),
            )
        )

        self.assertTrue(
            np.isclose(
                np.sum(average.first) / np.sum(average.norm),
                np.mean(a[np.logical_not(m)]),
            )
        )

        self.assertTrue(np.all(np.equal(average.norm, np.sum(np.logical_not(m), axis=0))))

    def test_static_add_point(self):
        """
        Add data point-by-point.
        """

        a = np.random.random(35 * 50).reshape(35, 50)
        m = np.random.random(35 * 50).reshape(35, 50) > 0.8
        average = enstat.static(shape=a.shape[1])

        for i in range(a.shape[0]):
            for j in np.argwhere(~m[i, :]).ravel():
                average.add_point(a[i, j], j)

        self.assertTrue(
            np.isclose(
                np.sum(average.first) / np.sum(average.norm),
                np.mean(a[np.logical_not(m)]),
            )
        )

        self.assertTrue(
            np.isclose(
                np.sum(average.first) / np.sum(average.norm),
                np.mean(a[np.logical_not(m)]),
            )
        )

        self.assertTrue(np.all(np.equal(average.norm, np.sum(np.logical_not(m), axis=0))))

    def test_dynamic1d(self):
        """
        Dynamically grow shape.
        """

        average = enstat.dynamic1d()

        average += np.array([1, 2, 3])
        average += np.array([1, 2, 3])
        average += np.array([1, 2])
        average += np.array([1])
        average += np.array([1, 2, 3, 4])
        average += np.array([1, 2, 3, 4])

        self.assertTrue(np.allclose(average.mean(), np.array([1, 2, 3, 4])))
        self.assertTrue(np.allclose(average.std(), np.array([0, 0, 0, 0])))
        self.assertEqual(average.shape, (4,))
        self.assertEqual(average.size, 4)


class Test_defaultdict(unittest.TestCase):
    """
    functionality
    """

    def test_scalar(self):
        average = defaultdict(enstat.scalar)

        a = np.random.random(50 * 20).reshape(50, 20)
        b = np.random.random(52 * 21).reshape(52, 21)

        for i in range(a.shape[0]):
            average["a"] += a[i, :]

        for i in range(b.shape[0]):
            average["b"] += b[i, :]

        self.assertTrue(np.isclose(average["a"].mean(), np.mean(a)))
        self.assertTrue(np.isclose(average["b"].mean(), np.mean(b)))

    def test_static(self):
        average = defaultdict(enstat.static)

        a = np.random.random(35 * 50 * 20).reshape(35, 50, 20)
        b = np.random.random(37 * 52 * 21).reshape(37, 52, 21)

        for i in range(a.shape[0]):
            average["a"] += a[i, :, :]

        for i in range(b.shape[0]):
            average["b"] += b[i, :, :]

        self.assertTrue(np.allclose(average["a"].mean(), np.mean(a, axis=0)))
        self.assertTrue(np.allclose(average["b"].mean(), np.mean(b, axis=0)))
        self.assertTrue(average["a"].shape == a.shape[1:])
        self.assertTrue(average["b"].shape == b.shape[1:])


class Test_restore(unittest.TestCase):
    """
    Restore existing average.
    """

    def test_scalar(self):
        data = np.random.random(100)
        average = enstat.scalar()

        for i in data:
            average += i

        restored = enstat.scalar.restore(**dict(average))

        self.assertTrue(np.isclose(average.first, restored.first))
        self.assertTrue(np.isclose(average.second, restored.second))
        self.assertTrue(average.norm == restored.norm)

        self.assertTrue(np.isclose(np.mean(data), average.mean()))
        self.assertTrue(np.isclose(np.std(data), average.std(), rtol=5e-1, atol=1e-3))

        self.assertTrue(np.isclose(np.mean(data), restored.mean()))
        self.assertTrue(np.isclose(np.std(data), restored.std(), rtol=5e-1, atol=1e-3))

    def test_static(self):
        data = np.random.random(31 * 50 * 11).reshape(31, 50, 11)
        average = enstat.static()

        for i in range(data.shape[0]):
            average += data[i, ...]

        restored = enstat.static.restore(**dict(average))

        self.assertTrue(np.allclose(average.first, restored.first))
        self.assertTrue(np.allclose(average.second, restored.second))
        self.assertTrue(np.all(average.norm == restored.norm))

        self.assertTrue(np.allclose(np.mean(data, axis=0), average.mean()))
        self.assertTrue(np.allclose(np.std(data, axis=0), average.std(), rtol=5e-1, atol=1e-3))

        self.assertTrue(np.allclose(np.mean(data, axis=0), restored.mean()))
        self.assertTrue(np.allclose(np.std(data, axis=0), restored.std(), rtol=5e-1, atol=1e-3))

    def test_dynamic1d(self):
        data = np.random.random(31 * 50).reshape(31, 50)
        average = enstat.dynamic1d()

        for i in data:
            average += i

        restored = enstat.dynamic1d.restore(**dict(average))

        self.assertTrue(np.allclose(average.first, restored.first))
        self.assertTrue(np.allclose(average.second, restored.second))
        self.assertTrue(np.all(average.norm == restored.norm))

        self.assertTrue(np.allclose(np.mean(data, axis=0), average.mean()))
        self.assertTrue(np.allclose(np.std(data, axis=0), average.std(), rtol=5e-1, atol=1e-3))

        self.assertTrue(np.allclose(np.mean(data, axis=0), restored.mean()))
        self.assertTrue(np.allclose(np.std(data, axis=0), restored.std(), rtol=5e-1, atol=1e-3))


if __name__ == "__main__":
    unittest.main()



from pathlib import Path

from setuptools import find_packages
from setuptools import setup

project_name = "enstat"

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name=project_name,
    license="MIT",
    author="Tom de Geus",
    author_email="tom@geus.me",
    description="Ensemble averages",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="Statistics, Ensemble",
    url=f"https://github.com/tdegeus/{project_name:s}",
    packages=find_packages(exclude=["tests"]),
    use_scm_version={"write_to": f"{project_name}/_version.py"},
    setup_requires=["setuptools_scm"],
    install_requires=["numpy"],
)



import warnings

import numpy as np


class Scalar:
    r"""
    Allocate the ensemble average of a scalar.
    Add samples to it using :py:func:`Scalar.add_sample`.
    The mean, variance, and standard deviation can be obtained at any time.
    Also the sums of the first and statistical moments, as well as the number of samples can be
    obtained at any time.

    Continue an old average by specifying:

    :param float first: Sum of the first moment.
    :param float second: Sum of the second moment.
    :param int norm: Number of samples.
    """

    def __init__(self, first=0, second=0, norm=0):
        warnings.warn("Deprecated. Use: enstat.scalar.", DeprecationWarning)

        self.m_first = first
        self.m_second = second
        self.m_norm = norm

    def add_sample(self, data):
        r"""
        Add a sample.
        Internally changes the sums of the first and second statistical moments and normalisation.

        :param float data: the scalar sample.
        """

        self.m_first += np.sum(data.ravel())
        self.m_second += np.sum(data.ravel() ** 2)
        self.m_norm += data.size

    def mean(self):
        r"""
        Obtain the current mean.
        N.B. samples can be added afterwards without any problems.
        """

        if self.m_norm == 0:
            return np.NaN

        return self.m_first / self.m_norm

    def variance(self):
        r"""
        Obtain the current variance.
        N.B. samples can be added afterwards without any problems.
        """

        if self.m_norm <= 1:
            return np.NaN

        return (
            (self.m_second / self.m_norm - (self.m_first / self.m_norm) ** 2)
            * self.m_norm
            / (self.m_norm - 1)
        )

    def std(self):
        r"""
        Obtain the current standard deviation.
        N.B. samples can be added afterwards without any problems.
        """

        return np.sqrt(self.variance())

    def norm(self):
        r"""
        Return normalisation: the number of samples.

        :return: Scalar.
        """

        return self.m_norm

    def first(self):
        r"""
        Sum of the first statistical moment.

        :return: Scalar.
        """

        return self.m_first

    def second(self):
        r"""
        Sum of the second statistical moment.

        :return: Scalar.
        """

        return self.m_second


class StaticNd:
    r"""
    Allocate the ensemble average of an Nd-array (of same size for all samples).
    Add samples to it using :py:func:`StaticNd.add_sample`.
    The mean, variance, and standard deviation can be obtained at any time.
    Also the sums of the first and statistical moments, as well as the number of samples can be
    obtained at any time.

    Continue an old average by specifying ``first``, ``second``, ``norm``.

    :param bool compute_variance:
        If set ``False`` no second moment will be computed.
        The variance an standard deviation will not be available.

    :param list shape, optional:
        The shape of the data.
        If not specified it is determined form the first sample.

    :param type dtype, optional:
        The numpy-type of the data.
        If not specified it is determined form the first sample.

    :param np.array<T> first: Continued computation: Sum of the first moment.
    :param np.array<T> second: Continued computation: Sum of the second moment.
    :param np.array<np.int64> norm: Continued computation: Number of samples.
    """

    def __init__(
        self,
        compute_variance=True,
        shape=None,
        dtype=None,
        first=None,
        second=None,
        norm=None,
    ):
        warnings.warn("Deprecated. Use: enstat.static.", DeprecationWarning)

        self.m_compute_variance = compute_variance
        self.m_first = first
        self.m_second = second
        self.m_norm = norm
        self.m_shape = shape
        self.m_dtype = dtype

    def _allocate(self, data):
        if self.m_first is not None:
            assert data.shape == self.m_shape
            return

        self.m_shape = self.m_shape if self.m_shape is not None else data.shape
        self.m_dtype = self.m_dtype if self.m_dtype is not None else data.dtype

        self.m_norm = np.zeros(self.m_shape, np.int64)
        self.m_first = np.zeros(self.m_shape, self.m_dtype)

        if self.m_compute_variance:
            self.m_second = np.zeros(self.m_shape, self.m_dtype)

    def shape(self):
        r"""
        Return the data's shape.
        """
        return self.m_shape

    def size(self):
        r"""
        Return the data's size.
        """
        return np.prod(self.m_shape)

    def add_sample(self, data, mask=None):
        r"""
        Add a sample.
        Internally changes the sums of the first and second statistical moments and normalisation.

        :param np.array<T> data: the sample.
        :param np.array<bool> mask: optional, mask entries.
        """

        self._allocate(data)

        # masked data
        if mask is not None:
            incl = np.logical_not(mask)
            self.m_norm[incl] += 1
            self.m_first[incl] += data[incl]
            if self.m_compute_variance:
                self.m_second[incl] += data[incl] ** 2
            return

        # unmasked data
        self.m_norm += 1
        self.m_first += data
        if self.m_compute_variance:
            self.m_second += data**2

    def mean(self):
        r"""
        Obtain the current mean.
        N.B. samples can be added afterwards without any problems.
        """

        if self.m_norm is None:
            return None

        n = np.where(self.m_norm > 0, self.m_norm, 1)
        ret = self.m_first / n
        return np.where(self.m_norm > 0, ret, np.NaN)

    def variance(self):
        r"""
        Obtain the current variance.
        N.B. samples can be added afterwards without any problems.
        """

        if self.m_norm is None:
            return None

        assert self.m_compute_variance
        n = np.where(self.m_norm > 1, self.m_norm, 2)
        ret = (self.m_second / n - (self.m_first / n) ** 2) * n / (n - 1)
        return np.where(self.m_norm > 1, ret, np.NaN)

    def std(self):
        r"""
        Obtain the current standard deviation.
        N.B. samples can be added afterwards without any problems.
        """

        if self.m_norm is None:
            return None

        return np.sqrt(self.variance())

    def norm(self):
        r"""
        Return normalisation: the number of samples.

        :return: Scalar.
        """

        return self.m_norm

    def first(self):
        r"""
        Sum of the first statistical moment.

        :return: Scalar.
        """

        return self.m_first

    def second(self):
        r"""
        Sum of the second statistical moment.

        :return: Scalar.
        """

        assert self.m_compute_variance
        return self.m_second


def _expand_array1d(data, size):
    tmp = np.zeros((size), data.dtype)
    tmp[: data.size] = data
    return tmp


class Dynamic1d(StaticNd):
    r"""
    Allocate the ensemble average of an 1d-array (which grows depending of the size of the samples).
    Add samples to it using :py:func:`Dynamic1d.add_sample`.
    The mean, variance, and standard deviation can be obtained at any time.
    Also the sums of the first and statistical moments, as well as the number of samples can be
    obtained at any time.

    Continue an old average by specifying ``first``, ``second``, ``norm``.

    :param bool compute_variance:
        If set ``False`` no second moment will be computed.
        The variance an standard deviation will not be available.

    :param list size, optional:
        The initial size of the data.

    :param type dtype, optional:
        The numpy-type of the data.
        If not specified it is determined form the first sample.

    :param np.array<T> first: Continued computation: Sum of the first moment.
    :param np.array<T> second: Continued computation: Sum of the second moment.
    :param np.array<np.int64> norm: Continued computation: Number of samples.
    """

    def __init__(
        self,
        compute_variance=True,
        size=None,
        dtype=None,
        first=None,
        second=None,
        norm=None,
    ):
        warnings.warn("Deprecated. Use: enstat.dynamic1d.", DeprecationWarning)

        self.m_compute_variance = compute_variance
        self.m_first = first
        self.m_second = second
        self.m_norm = norm
        self.m_size = size
        self.m_shape = [self.m_size]
        self.m_dtype = dtype

    def _allocate(self, data):
        r"""
        Allocate data if necessary.
        """

        if self.m_first is not None:
            return

        self.m_size = self.m_size if self.m_size is not None else data.size
        self.m_dtype = self.m_dtype if self.m_dtype is not None else data.dtype
        self.m_shape = [self.m_size]

        self.m_norm = np.zeros((self.m_size), np.int64)
        self.m_first = np.zeros((self.m_size), self.m_dtype)

        if self.m_compute_variance:
            self.m_second = np.zeros((self.m_size), self.m_dtype)

    def _expand(self, data):
        if data.size <= self.m_size:
            return

        self.m_size = data.size
        self.m_shape = [self.m_size]
        self.m_norm = _expand_array1d(self.m_norm, data.size)
        self.m_first = _expand_array1d(self.m_first, data.size)

        if self.m_compute_variance:
            self.m_second = _expand_array1d(self.m_second, data.size)

    def add_sample(self, data):
        r"""
        Add a sample.
        Internally changes the sums of the first and second statistical moments and normalisation.

        :param np.array<T> data: the sample.
        """

        assert data.ndim == 1

        self._allocate(data)
        self._expand(data)

        self.m_norm[: data.size] += 1
        self.m_first[: data.size] += data

        if self.m_compute_variance:
            self.m_second[: data.size] += data**2


if __name__ == "__main__":
    pass
